# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

import sys
from os.path import abspath, dirname, join

# mark sources root
sys.path.insert(0, join(abspath(dirname(__file__))))
